﻿Class LinkedList(Of ListType)

    Private Class Cell
        Public Value As ListType
        Public NextCell As Cell

        Sub New()
            MyBase.New()
        End Sub

        Sub New(ByVal Value As ListType, ByVal NextCell As Cell)
            Me.Value = Value
            Me.NextCell = NextCell
        End Sub

    End Class

    Dim FirstCell, LastCell As Cell
    Sub Add(ByVal Item As ListType)
        If pCount = 0 Then
            ' يجب إنشاء نسخة جديدة من خانة البداية وخانة النهاية
            FirstCell = New Cell
            LastCell = New Cell
            FirstCell.Value = Item
            LastCell = FirstCell
            pCount = 1
        Else
            Dim E As New Cell
            E.Value = Item
            LastCell.NextCell = E
            LastCell = E
            pCount += 1
        End If
    End Sub

    Dim pCount As Integer = 0
    ReadOnly Property Count() As Integer
        Get
            Return pCount
        End Get
    End Property

    Private Function GetCell(ByVal Index As Integer) As Cell
        Try
            If Index < 0 OrElse Index >= pCount Then
                Throw New Exception("رقم العنصر غير صالح")
            End If
            Dim C As Cell = FirstCell
            Dim I As Integer = 0
            Do Until I = Index
                C = C.NextCell
                I += 1
            Loop
            Return C
        Finally

        End Try
    End Function

    Default Property Item(ByVal Index As Integer) As ListType
        Get
            Try
                Return GetCell(Index).Value
            Finally

            End Try
        End Get
        Set(ByVal Value As ListType)
            Try
                GetCell(Index).Value = Value
            Finally

            End Try
        End Set
    End Property
    
    Sub AddArray(ByRef Items() As ListType)
        For I As Integer = 0 To UBound(Items)
            Add(Items(I))
        Next
    End Sub

    Sub AddList(ByRef NewList As LinkedList(Of ListType))
        If pCount = 0 Then
            FirstCell = NewList.FirstCell
            LastCell = NewList.LastCell
            pCount = NewList.pCount
        Else
            LastCell.NextCell = NewList.FirstCell
            LastCell = NewList.LastCell
            pCount += NewList.pCount
        End If
    End Sub

    Sub CopyList(ByRef NewList As LinkedList(Of ListType))
        Dim C As Cell = NewList.FirstCell
        Do Until C Is Nothing
            Me.Add(C.Value)
            C = C.NextCell
        Loop
    End Sub

    Function SubList(ByVal St As Integer) As LinkedList(Of ListType)
        Try
            Dim L As New LinkedList(Of ListType)
            L.FirstCell = GetCell(St)
            L.LastCell = LastCell
            L.pCount = pCount - St
            Return L
        Finally

        End Try
    End Function

    Function CopySubList(ByVal St As Integer, Optional ByVal Count _
                As Integer = -1) As LinkedList(Of ListType)
        Try
            If Count = -1 Then Count = pCount - St
            Dim I As Integer = 0
            Dim L As New LinkedList(Of ListType)
            Dim C As Cell = GetCell(St)
            Do Until C Is Nothing OrElse I = Count
                L.Add(C.Value)
                I += 1
                C = C.NextCell
            Loop
            Return L
        Finally

        End Try
    End Function

    Sub Insert(ByVal Item As ListType, ByVal Index As Integer)
        Try
            If Index = pCount Then
                Add(Item)
            Else
                Dim C As New Cell(Item, GetCell(Index))
                If Index = 0 Then
                    FirstCell = C
                Else : GetCell(Index - 1).NextCell = C
                End If
                pCount += 1
            End If
        Finally
        End Try
    End Sub

    Sub InsertList(ByRef NewList As LinkedList(Of ListType), _
             ByVal Index As Integer)
        Try
            If Index = pCount Then
                AddList(NewList)
            Else
                If Index = 0 Then
                    FirstCell = NewList.FirstCell
                Else : GetCell(Index - 1).NextCell = NewList.FirstCell
                End If
                NewList.LastCell = GetCell(Index)
                pCount += NewList.Count
            End If
        Finally
        End Try
    End Sub

    Sub CopyList(ByRef NewList As LinkedList(Of ListType), ByVal Index As Integer)
        Dim C As Cell = NewList.FirstCell
        Do Until C Is Nothing
            Insert(C.Value, Index)
            C = C.NextCell
        Loop
    End Sub

    Sub InsertArray(ByRef Items() As ListType, ByVal Index As Integer)
        Dim L As New LinkedList(Of ListType)
        L.AddArray(Items)
        InsertList(L, Index)
    End Sub

    Sub Clear()
        FirstCell = Nothing
        LastCell = Nothing
        pCount = 0
    End Sub

    Sub RemoveAt(ByVal Index As Integer)
        Try
            If Index = 0 Then ' حذف أول عنصر
                If pCount = 1 Then
                    ' العنصر الأول هو العنصر الأخير وسيتم حذفه
                    FirstCell = Nothing
                    LastCell = Nothing
                Else ' جعل العنصر الثاني هو العنصر الأول
                    FirstCell = FirstCell.NextCell
                End If
            ElseIf Index = pCount - 1 Then
                ' حذف آخر عنصر بجعل العنصر السابق له هو العنصر الأخير
                LastCell = GetCell(Index - 1)
                LastCell.NextCell = Nothing
            Else 'حذف العنصر الحالي، بجعل العنصر السابق يشير إلى العنصر التالي
                GetCell(Index - 1).NextCell = GetCell(Index + 1)
            End If
            pCount -= 1
        Finally

        End Try
    End Sub

    Sub RemoveLastItem()
        Try
            If pCount = 1 Then
                FirstCell = Nothing
                LastCell = Nothing
            Else
                LastCell = GetCell(pCount - 2)
                LastCell.NextCell = Nothing
            End If
            pCount -= 1
        Finally

        End Try
    End Sub

    Function RemoveRange(ByVal St As Integer, ByVal DelCount As Integer) As Integer
        Try
            Dim NxtCell As Integer = St + DelCount
            If NxtCell > pCount Then DelCount = pCount - St
            If St = 0 Then
                If pCount = DelCount Then
                    FirstCell = Nothing
                    LastCell = Nothing
                Else
                    FirstCell = GetCell(DelCount)
                End If
            ElseIf NxtCell = pCount Then
                LastCell = GetCell(St - 1)
                LastCell.NextCell = Nothing
            Else
                GetCell(St - 1).NextCell = GetCell(NxtCell)
            End If
            pCount -= DelCount
            Return DelCount
        Finally
        End Try
    End Function

    Dim pCurCell As Cell
    ReadOnly Property CurItem() As ListType
        Get
            Try
                If pCurCell Is Nothing Then
                    Throw New Exception("لا توجد خانة حالية")
                End If
                Return pCurCell.Value
            Finally

            End Try
        End Get
    End Property

    Function MoveFirst() As Boolean
        If FirstCell Is Nothing Then Return False
        pCurCell = FirstCell
        Return True
    End Function

    Function MoveNext() As Boolean
        If pCurCell Is Nothing Then
            Throw New Exception("لا توجد خانة حالية")
        End If
        If pCurCell Is LastCell Then Return False
        pCurCell = pCurCell.NextCell
        Return True
    End Function

    Function MoveTo(ByVal Index As Integer) As Boolean
        Try
            pCurCell = GetCell(Index)
            Return True
        Catch
            Return False
        End Try
    End Function

    Function MoveLast() As Boolean
        If LastCell Is Nothing Then Return False
        pCurCell = LastCell
        Return True
    End Function

    Property FirstItem() As ListType
        Get
            If FirstCell Is Nothing Then
                Throw New Exception("القائمة فارغة")
            End If
            Return FirstCell.Value
        End Get
        Set(ByVal value As ListType)
            If FirstCell Is Nothing Then
                Throw New Exception("القائمة فارغة")
            End If
            FirstCell.Value = value
        End Set
    End Property

    Property LastItem() As ListType
        Get
            If LastCell Is Nothing Then
                Throw New Exception("القائمة فارغة")
            End If
            Return LastCell.Value
        End Get
        Set(ByVal value As ListType)
            If LastCell Is Nothing Then
                Throw New Exception("القائمة فارغة")
            End If
            LastCell.Value = value
        End Set
    End Property

    Function ToArray(Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1) As ListType()
        Try
            If En = -1 OrElse En >= pCount Then En = pCount - 1
            Dim A(En - St) As ListType
            Dim C As Cell = GetCell(St)
            For I As Integer = 0 To En - St
                A(I) = C.Value
                C = C.NextCell
            Next
            Return A
        Finally
        End Try
    End Function

    Sub ShiftRight(Optional ByVal Times As Integer = 1)
        If FirstCell Is Nothing Then Return
        For I As Integer = 1 To Times
            LastCell.NextCell = FirstCell
            FirstCell = LastCell
            LastCell = GetCell(pCount - 1)
            LastCell.NextCell = Nothing
        Next
    End Sub

    Sub ShiftLeft(Optional ByVal Times As Integer = 1)
        If FirstCell Is Nothing Then Return
        For I As Integer = 1 To Times
            LastCell.NextCell = FirstCell
            LastCell = FirstCell
            FirstCell = FirstCell.NextCell
            LastCell.NextCell = Nothing
        Next
    End Sub

    Public Shared Operator =(ByVal List1 As LinkedList _
            (Of ListType), ByVal List2 As LinkedList(Of ListType)) As Boolean
        If List1 Is List2 Then Return True
        If List1.Count <> List2.Count Then Return False
        Dim C1 As Cell = List1.FirstCell
        Dim C2 As Cell = List2.FirstCell
        Do Until C1 Is Nothing
            If Not C1.Value.Equals(C2.Value) Then Return False
            C1 = C1.NextCell
            C2 = C2.NextCell
        Loop
        Return True
    End Operator

    Public Shared Operator <>(ByVal List1 As LinkedList(Of  ListType), ByVal List2 As LinkedList(Of ListType)) As Boolean
        Return Not (List1 = List2)
    End Operator

    Public Shared Operator >(ByVal List1 As LinkedList(Of  ListType), ByVal List2 As LinkedList(Of ListType)) As Boolean
        Return (List1.Count > List2.Count)
    End Operator

    Public Shared Operator <(ByVal List1 As LinkedList(Of ListType), ByVal List2 As LinkedList(Of ListType)) As Boolean
        Return (List1.Count < List2.Count)
    End Operator

    Public Shared Operator >=(ByVal List1 As LinkedList(Of ListType), ByVal List2 As LinkedList(Of ListType)) As Boolean
        Return (List1.Count >= List2.Count)
    End Operator

    Public Shared Operator <=(ByVal List1 As LinkedList(Of ListType), ByVal List2 As LinkedList(Of ListType)) As Boolean
        Return (List1.Count <= List2.Count)
    End Operator

    Public Shared Operator Not(ByVal L1 As LinkedList _
            (Of ListType)) As Boolean
        Return (L1.Count = 0)
    End Operator

    Private Shared Function CheckSubList(ByVal C1 As Cell, _
                 ByVal C2 As Cell) As Boolean
        C1 = C1.NextCell
        C2 = C2.NextCell
        Do Until C1 Is Nothing
            If Not C1.Value.Equals(C2.Value) Then Return False
            C1 = C1.NextCell
            C2 = C2.NextCell
        Loop
        Return True
    End Function

    Public Shared Operator Like(ByVal List1 As LinkedList(Of  _
            ListType), ByVal List2 As LinkedList(Of ListType)) _
            As Boolean
        Dim L1 As Integer = List1.Count
        Dim L2 As Integer = List2.Count
        If L1 > L2 Then Return False

        If List1 = List2 Then Return True

        Dim C1 As Cell = List1.FirstCell
        Dim C2 As Cell = List2.FirstCell

        Do Until C2 Is Nothing
            If C1.Value.Equals(C2.Value) Then
                If CheckSubList(C1, C2) Then Return True
            End If
            C2 = C2.NextCell
            ' مع كل حركة إلى عنصر تالي، يقل عدد العناصر المتبقة في القائمة
            L2 -= 1
            If L1 > L2 Then Return False
        Loop
        Return False
    End Operator

    Public Shared Operator &(ByVal List1 As LinkedList _
            (Of ListType), ByVal List2 As LinkedList _
            (Of ListType)) As LinkedList(Of ListType)
        Dim L As New LinkedList(Of ListType)
        L.CopyList(List1)
        L.CopyList(List2)
        Return L
    End Operator

    Public Shared Operator +(ByVal List1 As LinkedList _
            (Of ListType), ByVal List2 As LinkedList(Of ListType)) _
            As LinkedList(Of Double)
        Dim L As New LinkedList(Of Double)
        Dim C1, C2 As Cell
        If List1.Count < List2.Count Then
            C1 = List1.FirstCell
            C2 = List2.FirstCell
        Else
            C1 = List2.FirstCell
            C2 = List1.FirstCell
        End If
        Do Until C1 Is Nothing
            L.Add(CDbl(CObj(C1.Value)) + CDbl(CObj(C2.Value)))
            C1 = C1.NextCell : C2 = C2.NextCell
        Loop
        ' نقل العناصر الزائدة من القائمة الأطول
        Do Until C2 Is Nothing
            L.Add(CDbl(CObj(C2.Value)))
            C2 = C2.NextCell
        Loop
        Return L
    End Operator

    Public Shared Widening Operator CType(ByVal List1 As  _
            LinkedList(Of ListType)) As LinkedList(Of Double)
        Dim L As New LinkedList(Of Double)
        Dim C As Cell = List1.FirstCell
        Do Until C Is Nothing
            L.Add(CDbl(CObj(C.Value)))
            C = C.NextCell
        Loop
        Return L
    End Operator

    Public Shared Narrowing Operator CType(ByVal List1 As LinkedList(Of ListType)) As LinkedList(Of Integer)
        Dim L As New LinkedList(Of Integer)
        Dim C As Cell = List1.FirstCell
        Do Until C Is Nothing
            L.Add(CInt(CObj(C.Value)))
            C = C.NextCell
        Loop
        Return L
    End Operator

    Public Shared Operator >>(ByVal List1 As LinkedList(Of  _
            ListType), ByVal Times As Integer) As LinkedList(Of  _
            ListType)
        Dim L As New LinkedList(Of ListType)
        L.CopyList(List1)
        L.ShiftRight(Times)
        Return L
    End Operator

    Public Shared Operator <<(ByVal List1 As LinkedList(Of  _
            ListType), ByVal Times As Integer) As LinkedList(Of  _
            ListType)
        Dim L As New LinkedList(Of ListType)
        L.CopyList(List1)
        L.ShiftLeft(Times)
        Return L
    End Operator

    Public Shared Operator IsTrue(ByVal List1 As LinkedList(Of  _
            ListType)) As Boolean
        Return (List1.pCount > 0)
    End Operator

    Public Shared Operator IsFalse(ByVal List1 As LinkedList _
            (Of ListType)) As Boolean
        Return (List1.pCount = 0)
    End Operator

    Public Shared Widening Operator CType(ByVal List1 As LinkedList(Of ListType)) As Boolean
        Return (List1.pCount > 0)
    End Operator

End Class
