﻿Class List(Of ListType As IComparable)
    Dim S() As ListType

    Sub New(ByVal InitialCapacity As Integer)
        pCapacity = InitialCapacity
        If pCapacity < 1 Then pCapacity = 50
        ReDim S(pCapacity - 1)
        pCount = 0
    End Sub

    Dim pCount As Integer
    ReadOnly Property Count() As Integer
        Get
            Return pCount
        End Get
    End Property

    Private pCapacity As Integer
    Property Capacity() As Integer
        Get
            Return pCapacity
        End Get
        Set(ByVal Value As Integer)
            If pCapacity = Value Then Exit Property
            If Value < 1 Then
                Console.WriteLine("Bad Value")
                Exit Property
            End If
            pCapacity = Value
            ' سنغير حجم المصفوفة إلى السعة الجديدة، مع حفظ القيم القديمة
            ReDim Preserve S(pCapacity - 1)
            ' قد تكون السعة الجديدة أصغر من عدد العناصر
            ' ستضيع العناصر الزائدة، لهذا يجب أن نغير عدد العناصر أيضا
            If pCapacity < pCount Then pCount = pCapacity
        End Set
    End Property

    Default Property Item(ByVal Index As Integer) As ListType
        Get
            Return S(Index)
        End Get
        Set(ByVal value As ListType)
            S(Index) = value
        End Set
    End Property

    Enum SortState
        ASC = 0
        DESC = 1
        None = 2
    End Enum

    Sub QuickSort(ByRef Arr() As ListType, Optional ByVal SortDirection As Integer = 0, Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1)
        If En = -1 Then En = UBound(Arr)
        Dim I As Integer, LAC As Integer = St, Temp As ListType
        For I = St + 1 To En
            If (SortDirection = 1) Xor (Arr(I).CompareTo(Arr(St)) < 1) Then
                LAC += 1
                ' تبادل قيمتي الخانتين، إذا كانتا بالفعل خانتين مختلفتين
                If LAC <> I Then
                    Temp = Arr(LAC)
                    Arr(LAC) = Arr(I)
                    Arr(I) = Temp
                End If
            End If
        Next
        ' تبادل قيمتي الخانة المرجعيّة مع آخر أصغر رقم
        If LAC <> St Then
            Temp = Arr(LAC)
            Arr(LAC) = Arr(St)
            Arr(St) = Temp
        End If
        If LAC - St > 1 Then
            QuickSort(Arr, SortDirection, St, LAC - 1)
        End If
        If En - LAC > 1 Then
            QuickSort(Arr, SortDirection, LAC + 1, En)
        End If
    End Sub

    Sub Sort(Optional ByVal State As SortState = SortState.ASC, Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1)
        pSorted = State
        If En = -1 Then En = pCount - 1
        If State <> SortState.None Then QuickSort(S, State, St, En)
    End Sub

    Private pSorted As SortState = SortState.None
    Property Sorted() As SortState
        Get
            Return pSorted
        End Get
        Set(ByVal value As SortState)
            If pSorted <> value Then Sort(value)
            ' لا نحتاج إلى الجملة التالية، لأن إجراء الترتيب سيقوم بها
            ' pSorted = value
        End Set
    End Property

    Function BinarySearch(ByRef Arr() As ListType, ByVal Target _
                 As ListType, Optional ByVal SortDirection As Integer = 0, _
                 Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1) As Integer
        If En = -1 Then En = UBound(Arr)
        ' شرط توقف: العثور على عنصر البحث في خانة البداية أو النهاية
        If Arr(St).Equals(Target) Then Return St
        If Arr(En).Equals(Target) Then Return En
        Dim Mdl As Integer = (St + En + 1) \ 2
        If Arr(Mdl).Equals(Target) Then
            Return Mdl
        ElseIf (SortDirection = 1) Xor (Arr(Mdl).CompareTo(Target) = -1) Then
            If En - Mdl < 2 Then Return -(Mdl + 1)
            ' البحث في الجزء الأخير من المصفوفة الحاليّة
            St = Mdl + 1
            En -= 1
            ' نتيجة البحث هي
            ' الناتجة عن البحث في الجزء الأخير من المصفوفة
            Return BinarySearch(Arr, Target, SortDirection, St, En)
        Else
            If Mdl - St < 2 Then Return -Mdl
            ' البحث في الجزء الأوّل من المصفوفة الحاليّة
            St += 1
            En = Mdl - 1
            ' نتيجة البحث هي
            '  تلك الناتجة عن البحث في النصف الأوّل من المصفوفة
            Return BinarySearch(Arr, Target, SortDirection, St, En)
        End If
    End Function

    Function Add(ByVal Item As ListType) As Integer
        Return Insert(Item, pCount)
    End Function

    Function Insert(ByVal Item As ListType, ByVal Index As Integer) As Integer
        pCount += 1
        If pCount > pCapacity Then
            pCapacity += 50
            ReDim Preserve S(pCapacity - 1)
        End If

        If pSorted <> SortState.None Then
            Index = BinarySearch(S, Item, pSorted, 0, pCount - 2)
        End If

        If Index < 0 Then Index = -Index

        For I As Integer = pCount - 1 To Index + 1 Step -1
            S(I) = S(I - 1)
        Next
        S(Index) = Item
        Return Index
    End Function

    Sub AddArray(ByRef A() As ListType)
        Dim Temp As SortState = pSorted
        pSorted = SortState.None
        For I As Integer = 0 To UBound(A)
            Insert(A(I), pCount)
            ' pCount لاحظ أن عدد العناصر سيزيد بعد كل إضافة.. أي أن المتغير 
            ' قيمته تزيد باستمرار في كل لفة داخل جملة التكرار.
        Next
        pSorted = Temp
    End Sub

    Sub InsertArray(ByRef A() As ListType, ByVal Index As Integer)
        ' لو كانت القائمة مرتبة، فسندرج المصفوفة في آخرها
        If pSorted <> SortState.None Then Index = pCount

        Dim L As Integer = UBound(A) + 1
        pCount += L
        If pCount > pCapacity Then
            pCapacity += 50 + L
            ReDim Preserve S(pCapacity - 1)
        End If

        ' زحزحة عناصر القائمة بمقدار عدد خانات المصفوفة المضافة
        For I As Integer = pCount - 1 To Index + L Step -1
            S(I) = S(I - L)
        Next
        ' إدراج خانات المصفوفة المضافة
        For I As Integer = 0 To L - 1
            S(Index + I) = A(I)
        Next
        ' استدعاء إجراء الترتيب سيرتب القائمة على حسب قيمة خاصية الترتيب
        Sort(pSorted)
    End Sub

    Function IndexOf(ByVal Item As ListType, Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1) As Integer
        If En = -1 Then En = pCount - 1
        If pSorted = SortState.None Then
            For I As Integer = St To En
                If S(I).Equals(Item) Then Return I
            Next
            Return -1
        Else
            Dim R As Integer = BinarySearch(S, Item, pSorted, St, En)
            ' لاحظ أننا لا نستطيع إعادة أرقام سالبة غير -1، 
            ' لأن القائمة ليست دائما مرتبة  وهذا ليس ناتج بحث ثنائي دائما.
            ' لهذا سنفحص القيمة العائدة من إجراء البحث الثنائي،
            ' ونحول القيم السالبة كلها إلى -1 قبل إعادتها إلى المستخدم.
            If R < 0 Then R = -1
            Return R
        End If
    End Function

    Function LastIndexOf(ByVal Item As ListType, Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1) As Integer
        If En = -1 Then En = pCount - 1
        If pSorted = SortState.None Then
            For I As Integer = En To St Step -1
                If S(I).Equals(Item) Then Return I
            Next
            Return -1
        Else
            Dim R As Integer = BinarySearch(S, Item, pSorted, St, En)
            If R < 0 Then Return -1
            Do
                ' لو لم يكن عنصر البحث موجودا في الخانة التالية، فغادر الإجراء
                If Not S(R + 1).Equals(Item) Then Return R
                ' عنصر البحث موجود في الخانة التالية..
                ' أزد النتيجة بواحد وواصل الفحص
                R += 1
            Loop
        End If
    End Function

    Function RemoveAt(ByVal Index As Integer) As Integer
        If Index < 0 OrElse Index >= pCount Then Return -1
        pCount -= 1
        For I As Integer = Index To pCount - 1
            S(I) = S(I + 1)
        Next
        Return Index
    End Function

    Function RemoveRange(ByVal Index As Integer, ByVal DeletedCount As _
                Integer) As Integer
        If Index < 0 OrElse Index >= pCount Then Return 0
        If Index + DeletedCount > pCount Then DeletedCount = pCount - Index

        pCount -= DeletedCount
        ' زحزحة خانات القائمة إلى أعلى
        For I As Integer = Index To pCount - DeletedCount
            S(I) = S(I + DeletedCount)
        Next
        Return DeletedCount
    End Function

    Function Remove(ByVal Item As ListType) As Integer
        Return RemoveAt(IndexOf(Item))
    End Function

    Sub Remove(ByVal ParamArray A() As ListType)
        For I As Integer = 0 To UBound(A)
            RemoveAt(IndexOf(A(I)))
        Next
    End Sub

    Function RemoveAll(ByVal Item As ListType) As Integer
        Dim C As Integer = 0
        Dim R As Integer
        Do
            R = IndexOf(Item)
            If R < 0 Then Return C
            RemoveAt(R)
            C += 1
        Loop
    End Function

    Sub Clear()
        pCount = 0
    End Sub

    Function ToArray(Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1) As ListType()
        If En = -1 OrElse En >= pCount Then En = pCount - 1
        Dim A(En - St) As ListType
        For I As Integer = St To En
            A(I - St) = S(I)
        Next
        Return A
    End Function

End Class

