﻿Public Class MemoryStream : Inherits Stream
    Dim M() As Byte, pCapacity As Long = 255
    Dim pLength As Long, pPosition As Long

    Sub New()
        ReDim M(pCapacity - 1) : pLength = 0 : pPosition = 0
    End Sub

    Sub New(ByVal Capacity As Long)
        pCapacity = Capacity : ReDim M(pCapacity - 1)
        pLength = 0 : pPosition = 0
    End Sub

    Sub New(ByVal Data() As Byte, ByVal CanWrite As _
            Boolean, Optional ByVal St As Long = 0, _
            Optional ByVal En As Long = -1)
        pCanWrite = CanWrite
        If En = -1 Then En = UBound(Data)
        pLength = En - St + 1
        If pCanWrite Then
            If pLength > pCapacity Then
                pCapacity += 255 : ReDim Preserve M(pCapacity - 1)
            End If
        Else : pCapacity = pLength
        End If
        ReDim M(pCapacity - 1)
        For I As Integer = St To En
            M(I) = Data(I)
        Next
        pPosition = 0
    End Sub
    Public Overrides Sub Close()
        Erase M ' محو المصفوفة
        pPosition = -1
        pLength = -1
    End Sub

    Public Overrides Function ReadByte() As Integer
        If pPosition >= pLength Then Return -1
        ReadByte = M(pPosition)
        pPosition += 1
    End Function

    Public Overrides Function Read(ByRef Data() As Byte, _
                Optional ByVal St As Integer = 0, _
                Optional ByVal Count As Integer = 1) As Integer
        If pPosition + Count - 1 >= pLength Then Return -1
        For I As Integer = 0 To Count - 1
            Data(St + I) = M(pPosition)
            pPosition += 1
        Next
    End Function

    Public Overrides Sub WriteByte(ByVal value As Byte)
        If pCanWrite = False Then
            Throw New Exception("هذه البيانات للقراءة فقط")
        End If
        If pPosition >= pLength Then
            pLength += 1
            If pLength > pCapacity Then
                pCapacity += 255
                ReDim Preserve M(pCapacity - 1)
            End If
        End If
        M(pPosition) = value
        pPosition += 1
    End Sub

    Public Overrides Sub Write(ByVal Data() As Byte, _
                Optional ByVal St As Integer = 0, _
                Optional ByVal Count As Integer = 1)
        If pCanWrite = False Then
            Throw New Exception("هذه البيانات للقراءة فقط")
        End If
        If pPosition + Count - 1 >= pLength Then
            pLength = pPosition + Count
            If pLength > pCapacity Then
                pCapacity += 255
                ReDim Preserve M(pCapacity - 1)
            End If
        End If
        For I As Integer = St To St + Count - 1
            M(pPosition) = Data(I)
            pPosition += 1
        Next
    End Sub

    Public Overrides Function Seek(ByVal Offset As Long, _
                Optional ByVal origin As SeekOrigin = SeekOrigin.Current) As Long
        Dim Pos As Long
        Select Case origin
            Case SeekOrigin.Begin
                Pos = Offset
            Case SeekOrigin.Current
                Pos = pPosition + Offset
            Case Else 'SeekOrigin.End 
                Pos = pLength - 1 + Offset
        End Select
        If Pos >= pLength OrElse Pos < 0 Then Return -1
        pPosition = Pos
        Return Pos
    End Function

    Public Overrides ReadOnly Property CanRead() As Boolean
        Get
            Return pCanRead
        End Get
    End Property

    Public Overrides ReadOnly Property CanWrite() As Boolean
        Get
            Return pCanWrite
        End Get
    End Property

    Public Overrides ReadOnly Property Length() As Long
        Get
            Return pLength
        End Get
    End Property

    Public Overrides Property Position() As Long
        Get
            Return pPosition
        End Get
        Set(ByVal Pos As Long)
            If Pos >= pLength OrElse Pos < 0 Then
                Throw New Exception("لا يمكن الانتقال إلى هذا الموضع")
            End If
            pPosition = Pos
        End Set
    End Property

    Public Overridable Property Capacity() As Long
        Get
            Return pCapacity
        End Get
        Set(ByVal value As Long)
            pCapacity = value
            ReDim Preserve M(pCapacity - 1)
        End Set
    End Property

    Public Function GetBuffer() As Byte()
        Return M
    End Function

    Public Function ToArray() As Byte()
        Dim A(pLength - 1) As Byte
        For I As Integer = 0 To pLength - 1
            A(I) = M(I)
        Next
        Return A
    End Function

    Public Sub WriteTo(ByVal S As Stream, Optional ByVal St As _
            Integer = 0, Optional ByVal En As Integer = -1)
        If En = -1 Then En = pLength - 1
        S.Write(M, St, En - St + 1)
    End Sub

End Class
