﻿Imports VB = Microsoft.VisualBasic

Public Class FileRecord(Of Rtype As Structure)
    Inherits Record(Of Rtype)
    Dim pFileName As String

    Function FileExists(ByVal FileName As String) As Boolean
        Try
            Dim L As Long = FileLen(FileName)
            Return True
        Catch ex As Exception
            Return False
        End Try
    End Function

    Enum FileMode
        Append
        Create
        CreateNew
        Open
        OpenOrCreate
        Truncate
    End Enum

    Private Sub CheckFile(ByVal mode As FileMode)
        Dim Exists As Boolean = FileExists(pFileName)
        Select Case mode
            Case FileMode.CreateNew
                If Exists Then
                    Throw New Exception("الملف موجود بالفعل")
                End If
            Case FileMode.Create, FileMode.Truncate
                ' محو محتويات الملف بمحو الملف نفسه
                Kill(pFileName)
            Case FileMode.Open
                If Not Exists Then
                    Throw New Exception("الملف غير موجود")
                End If
        End Select
    End Sub

    Dim No As Integer = FreeFile()

    Public Sub New(ByVal FileName As String, Optional ByVal _
                mode As FileMode = FileMode.OpenOrCreate, _
                Optional ByVal access As OpenAccess = _
                OpenAccess.ReadWrite, Optional ByVal share As  _
                OpenShare = OpenShare.LockWrite)
        pFileName = FileName
        CheckFile(mode)
        FileOpen(No, FileName, OpenMode.Random, access, share, RecordLength)
        If FileMode.Append Then VB.Seek(No, LOF(No) + 1)
        pCanWrite = Not (access = OpenAccess.Read)
        pCanRead = Not (access = OpenAccess.Write)
    End Sub

    Public Overrides Sub Close()
        FileClose(No)
    End Sub

    Public Overrides Function ReadRecord(Optional ByVal RecNo As Long = -1) As Rtype
        If pCanRead = False Then
            Throw New Exception("هذا الملف مفتوح للكتابة فقط")
        End If
        If RecNo > -1 Then VB.Seek(No, RecNo + 1)
        If EOF(No) Then Return Nothing
        Dim R As Rtype : FileGet(No, R) : Return R
    End Function

    Public Overrides Sub Read(ByRef Rec() As Rtype, Optional _
                ByVal RecNo As Long = -1, Optional ByVal St As _
               Integer = 0, Optional ByVal Count As Integer = 1)
        If pCanRead = False Then
            Throw New Exception("هذا الملف مفتوح للكتابة فقط")
        End If
        If RecNo > -1 Then VB.Seek(No, RecNo + 1)
        If VB.Seek(No) - 1 + Count > LOF(No) Then
            Count = LOF(No) - VB.Seek(No) + 1
            If Count < 1 Then Return
        End If
        Dim R As Rtype
        For I As Integer = 0 To Count - 1
            FileGet(No, R) : Rec(St + I) = R
        Next
    End Sub

    Public Overrides Sub WriteRecord(ByVal Rec As Rtype, _
                Optional ByVal RecNo As Long = -1)
        If pCanWrite = False Then
            Throw New Exception("هذا الملف مفتوح للقراءة فقط")
        End If
        If RecNo > -1 Then
            FilePut(No, Rec, RecNo + 1)
        Else
            FilePut(No, Rec)
        End If
    End Sub

    Public Overrides Sub Write(ByVal Rec() As Rtype, _
                Optional ByVal RecNo As Long = -1, _
                Optional ByVal St As Integer = 0, _
                Optional ByVal Count As Integer = 1)
        If pCanWrite = False Then
            Throw New Exception("هذا الملف مفتوح للقراءة فقط")
        End If
        If RecNo > -1 Then VB.Seek(No, RecNo + 1)
        For I As Integer = St To St + Count - 1
            FilePut(No, Rec(I))
        Next
    End Sub

    Public Overrides ReadOnly Property CanRead() As Boolean
        Get
            Return pCanRead
        End Get
    End Property

    Public Overrides ReadOnly Property CanWrite() As Boolean
        Get
            Return pCanWrite
        End Get
    End Property

    Public Overrides ReadOnly Property Length() As Long
        Get
            Return LOF(No)
        End Get
    End Property

    Public Overrides Property Position() As Long
        Get
            Return VB.Seek(No) - 1
        End Get
        Set(ByVal Pos As Long)
            If Pos >= LOF(No) OrElse Pos < 0 Then
                Throw New Exception("لا يمكن الانتقال إلى هذا الموضع")
            End If
            VB.Seek(No, Pos + 1)
        End Set
    End Property

    Public Function ToArray() As Rtype()
        Dim Pos As Long = VB.Seek(No)
        VB.Seek(No, 1)
        Dim R(LOF(No) - 1) As Rtype
        FileGet(No, R)
        VB.Seek(No, Pos)
        Return R
    End Function

    Public Overridable Sub WriteTo(ByVal Rec As Record(Of Rtype),
                Optional ByVal St As Integer = 0, Optional ByVal En As Integer = -1)
        If En = -1 Then En = LOF(No) - 1
        Dim Pos As Long = VB.Seek(No)
        VB.Seek(No, St + 1)
        Dim R As Rtype
        For I As Integer = 0 To En - St
            FileGet(No, R)
            Rec.WriteRecord(R)
        Next
        VB.Seek(No, Pos)
    End Sub

    ReadOnly Property FileName() As String
        Get
            Return pFileName
        End Get
    End Property

    Public Sub Lock(ByVal Position As Long, ByVal Length As Long)
        VB.Lock(No, Position + 1, Position + Length)
    End Sub

    Public Sub Unlock(ByVal Position As Long, ByVal Length As Long)
        VB.Unlock(No, Position + 1, Position + Length)
    End Sub

    Function GetFileOpenAttr() As OpenMode
        Return FileAttr(No)
    End Function
End Class
